#include <ncurses/curses.h>
#include <linux/soundcard.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>

#define _DEFINE_GLOBALS_
#define LCHAR (196)
#define clrscr() erase()
#define strcmpi strcasecmp

/* #define DEBUGLEVEL 4 */
/* #define version_string "LYN v0.8a.4L" */

#ifdef VOLUME_USE_MIXER
#define CD_MIXER_DEV (SOUND_MIXER_CD)
#define SETVOL(l,r) mix_set_volume(CD_MIXER_DEV,&(l),&(r))
#define GETVOL(l,r) mix_get_volume(CD_MIXER_DEV,&(l),&(r))
#else
#define SETVOL(l,r) set_vol(l,r);
#define GETVOL(l,r) cdv_getvol(l,r)
#endif


#include "lyn.h"
#include "vars.h"

playinfo _cd;
unsigned char minutes, seconds, frames;
unsigned long disklen,loopctr=0,lastrand;
double lastid=0;

short looped_track=-1;
unsigned short curcheck, curmark;

/* cd_datarec dd; */
cd_playlist pl;
#if DEBUGLEVEL
  int no_disc;
#endif
char *cur_trackname;		/* Take a guess */
/*int cur_index = 0;
int cur_frame;*/
struct play *playlist = NULL;
/*int cur_track = -1;*/
char *cur_artist;		/* Name of current CD's artist */
char cur_avoid;			/* Avoid flag */
char cur_contd;			/* Continued flag */
char *cur_cdname;		/* Album name */
int cur_nsections;		/* Number of sections currently defined */
int exit_on_eject = 0;
int cur_balance=10, info_modified;

int hascolor=0;

int main(int argc, char *argv[])
{
  char s[256];
  int i;
  int cccur=0;

#if DEBUGLEVEL
  printf("lyn: initializing\n");
  printf("* gdb: starting as %s, PID %d",argv[0],getpid());
  getchar();
#endif
  playing=curses_ready=0;
  read_config(&config);
  cccur=cd_status();
#if DEBUGLEVEL>3
  printf("lyn: post-status\n");
#endif
  menunum=1;
  ccc=0;
  listmode=0;
  oldstatus=0;
  read_volumes();
  bal=10;
/*   cd = &thiscd; */
  signal(SIGINT, SIG_IGN);
  signal(SIGWINCH,sig_winch_handler);
  if (cccur==1 || cccur==2)
    diskin=1;
  else if (!cccur || cccur==3)
    diskin=0;
  if (cur_cdmode==1)
    playing=1;
  else if (!cur_cdmode || cur_cdmode>=4)
    playing=0;
  else if (cur_cdmode==3)
    playing=1;
  if (!playing)
    cur_track=1;
  if (argc>1) {
    /* commandline stuff */
    _fast=1;
    printf(charstr(79,LCHAR));
    printf("\r\n");
    printf(version_string);
    printf(charstr(42-strlen(version_string),32));
    printf("%s\r",copyright_string);
    if (!strcmpi(argv[1],"play") || !strcmpi(argv[1],"pl")) {
      if (atoi(argv[2]))
	play_cd(atoi(argv[2]),0,thiscd.ntracks);
      else
	play_cd(1,0,thiscd.ntracks);
      printf("%s :: PLAYING",version_string);
      if (atoi(argv[2]))
        printf(" #%d\n",atoi(argv[2]));
      else
        printf("\n");
    } else if (!strcmpi(argv[1],"stop") || !strcmpi(argv[1],"st")) {
      stop_cd();
      printf("%s :: STOPPED\n",version_string);
    } else if (!strcmpi(argv[1],"eject") || !strcmpi(argv[1],"ej")) {
/*      cd_lock(UNLOCK);*/
      eject_cd(EJECT_TRAY);
      printf("%s :: EJECTED\n",version_string);
    } else if (!strcmpi(argv[1],"pause") || !strcmpi(argv[1],"pa")) {
      pause_cd();
      printf("%s :: PAUSED\n",version_string);
    } else if (!strcmpi(argv[1],"resume") || !strcmpi(argv[1],"re")) {
      pause_cd();
      printf("%s :: RESUMED\n",version_string);
    }/* else if (!strcmpi(argv[1],"close") || !strcmpi(argv[1],"cl")) {
	  cd_cmd(CLOSE_TRAY);
      cwrite(strlen(version_string)+2,2,7,":: |15CLOSED");
    }*/ else if (!strcmpi(argv[1],"mute") || !strcmpi(argv[1],"mu")) {
      cd_volume(0,10,25);
      printf("%s :: MUTED\n",version_string);
    } else if (!strcmpi(argv[1],"full") || !strcmpi(argv[1],"loud") ||
	       !strcmpi(argv[1],"fvol")) {
      cd_volume(25,10,25);
      printf("%s :: FULL VOLUME\n",version_string);
    }/* else if (!strcmpi(argv[1],"lock") || !strcmpi(argv[1],"lo")) {
      cd_lock(LOCK);
      cwrite(strlen(version_string)+2,2,7,":: |15LOCKED");
    } else if (!strcmpi(argv[1],"unlock") || !strcmpi(argv[1],"un")) {
      cd_lock(UNLOCK);
      cwrite(strlen(version_string)+2,2,7,":: |15UNLOCKED");
    }*/ else if ((!strcmpi(argv[1],"volume") || !strcmpi(argv[1],"vol")) &&
	       atoi(argv[2])) {
      SETVOL(atoi(argv[2]),atoi(argv[2]));
      printf("%s :: VOLUME == %.1f%%\n",version_string,(float)(atoi(argv[2])/255.)*100.);
    } else {
      printf("%s :: HELP\n\n",version_string);
      printf("%sFormat: LYN [ play [track] | stop | pause | resume* |\n",charstr(7,32));
      printf("%s              eject | mute* | full* | vol <1-255> ]\n\n",charstr(7,32));
      printf("%sRun LYN without arguments for interactive mode.",charstr(7,32));
      printf("\n\n\n\n\n");
      return 0;
    }
    printf("\n\n");
    return 0;
  }
#if DEBUGLEVEL>3
  printf("lyn: commandline done\n");
#endif  
  init_curses();
#if DEBUGLEVEL>3
  printf("lyn: curses init done\n");
#endif
  curses_ready=1;
#if DEBUGLEVEL>4
/*  test_toc(); 
  nodelay(stdscr,FALSE);
  getstr(s);
  nodelay(stdscr,TRUE);  
*/
#endif
  prep_new_config();

#if 0
  cwrite(1,2,10,"(for gdb)|7 running as %s, PID %d",argv[0],getpid());
  nodelay(stdscr,FALSE);
  getstr(s);
  nodelay(stdscr,TRUE);  
#endif  
  if (config.tracksumstart==_CFG_TRACKSUMSTART_SL)
    listmode++;
  else if (config.tracksumstart==_CFG_TRACKSUMSTART_DEBUG) {
    listmode=0;
    ccc=1;
  }
  if (config.forcevolume && config.forcevolume!=_CFG_FORCEVOLUME_NONE) {
    vol=config.forcevolume;
    SETVOL(vol,vol);
/*    cd_volume(vol,10,255);*/
  }
#if DEBUGLEVEL>3
  printf("lyn: config enforce done\n");
#endif
  gotoxy(1,1);
  setup_screen(ccc);
  paused=0;
  done=0;
/*  memset(&thiscd,0,sizeof(struct cdinfo)); */
/*  memset(&dd,0,sizeof(cd_datarec)); */
/*
  strcpy(sourcepath,argv[0]);
  if (strchr(sourcepath,'/'))
    for (i=strlen(sourcepath)-1; i>0; i--) {
      if (sourcepath[i]=='/') {
        sourcepath[i+1]=0;
        break;
      }
    }
  else
    sourcepath[0]=0;
  */
  strcpy(sourcepath,config.cddbpath);
  if (diskin)
    newdisk();
  else
    nodisk();
  if (!playing) {
    cur_track=1;
  } else {
    cd_status();
    where_the_fuck_are_we();
    highlight_track(1,cur_track);
  }
  highlight_track(1,cur_track);  
  lasttrack=cur_track;

  /* The main cycle */
  
  while (!done) {
    loopctr++;
    ts=cd_status();
    update_topline();
    expire_err();
    if (oldstatus!=ts) {
      if (!ts || ts==3 || cur_cdmode==5) {
        nodisk();
      } else if (ts==2) {
        report_status("LOGGING");
        newdisk();
      } else if (!cur_cdmode || cur_cdmode==4) {
        playing=0;
        highlight_track(cur_track,0);
      }
      show_status();
      oldstatus=ts;   
    }
    if (disklen!=thiscd.length) {
      newdisk();
      oldstatus=ts;
    }
    if (playmode==_PLAYMODE_TRACKLOOP && cur_track!=looped_track && looped_track!=-1) {
      play_song(looped_track);
      highlight_track(1,looped_track);
    } else if (playmode==_PLAYMODE_SHUFFLE &&
        cur_track!=lasttrack && playing && (loopctr-50)<=lastrand) {
      lasttrack=cur_track;
      lastrand=loopctr;
    } else if (playmode==_PLAYMODE_SHUFFLE &&
      cur_track!=lasttrack) {
      play_song(-2);
      lasttrack=cur_track;
      highlight_track(1,cur_track);
    }
    if (!update_volumes())
      drawvol();
    ch=getch();
    if (ch!=-1) {
/*      if (ch>='a' && ch<='z')
        ch-=32;*/
#if DEBUGLEVEL>4
      if (ch!=255)
        dwrite(20,LINES,15,"c %d/%o (%c)  ",ch,ch,ch);
#endif
      switch (ch) {
        case KEY_F(1):
          ccc=0;
          if (listmode==_LISTMODE_EXT) {
            for (listmode=0; listmode<21; listmode++)
              f_write(27,4+listmode,8,charstr(54,32));
            listmode=_LISTMODE_NORM;
            cstr(27,4,8,24);
            cstr(27,LINES-1,8,24);
	    cstr(61,4,8,20);
            f_write(61,5,15,"TRACK SUMMARY");
            f_write(64,6,15,"##  Len    Type");
            update_topline();
            draw_tracklist();
            highlight_track(1,cur_track);
          } else if (menunum++==1)
            menunum=0;
          draw_menu(menunum);
          break;
        case KEY_F(2):
          /* fortunately, we have only two fkeys, as this KEY_F(n) format
             is acutely depressing - might have meant some typing, but
             KEY_Fn seems more sensible notation if you're not doing loop
             comparisons or whatever, and for that F(n) could be used.
             Oh well. */
          if (!listmode)
            listmode++;
          else {
            for (listmode=0; listmode<LINES-4; listmode++)
              f_write(27,4+listmode,8,charstr(54,' '));
            cstr(27,4,8,24);
            cstr(27,LINES,8,24);
            cstr(61,4,8,20);
            f_write(61,5,15,"TRACK SUMMARY");
            f_write(64,6,15,"##  Len    Type");
            listmode=0;
            update_topline();
          }
          if (listmode==_LISTMODE_NORM) {
            draw_menu(menunum);
            if (playing)
              sprintf(s,"%s, %s",thiscd.title,thiscd.songtitle[cur_track-1]);
            else
              strcpy(s,dd.title);
            report_topline(15,s);
          } else {
            ccc=0;
            report_topline(8,copyright_string);
          }
          draw_tracklist();
          highlight_track(1,cur_track);
          break;          
	case KEY_F(3):
	  pl.used=!pl.used;
	  draw_tracklist();
	  break;
        case KEY_UP:
          if (vol>90)
            vol=100;
          else
            vol+=10;
          SETVOL(vol,vol);
/*          cd_volume(vol,bal,255);*/
          DRAWVOL();
          break;
        case KEY_DOWN: 
          if (vol<10)
            vol=0;
          else
            vol-=10;
          
          /*cd_volume(vol,bal,255);*/
          SETVOL(vol,vol);
          DRAWVOL();
	  break;
       /* note: volume controls don't work with current hw control code */
	case KEY_LEFT:
	  if (bal)
	    cd_volume(vol,--bal,255);
	  DRAWVOL();
	case KEY_RIGHT:
	  if (bal<20)
	    cd_volume(vol,++bal,255);
	  break;          
    	case 'p':
	  if (playing)
 	    pause_song();
	  else if (playmode==_PLAYMODE_SHUFFLE)
	    play_song(-2);
	  else if (!playing)
	    play_song(1);
	  else
	    play_song(-1);
          break;
        case 'z':
          read_dbase();
/*          read_cddb_dbase(thiscd.cddb_id,&dd);
          oldstatus=-1;
          update_topline();
          draw_tracklist();
          highlight_track(1,cur_track);          
          update_xtermtitle(playing,(cur_pos_rel-(cur_pos_rel % 60))/60,cur_pos_rel % 60);
*/
          break;
        case 'y':
          if (listmode!=_LISTMODE_EXT) {
            listmode=_LISTMODE_EXT;
            report_topline(8,copyright_string);
            draw_tracklist();
            highlight_track(1,cur_track);
          }
          edit_songdata();
          break;
	case 's':
	  stop_cd();
          f_write(19,16,5," ");
          playing=0;
          highlight_track(0,cur_track);
          f_write(19,7,13,"   ");
	  f_write(19,8,13,"      ");
	  f_write(19,9,13,"      ");
	  report_status("STOPPED");
	  break;
	case '[':
          play_song((playmode==_PLAYMODE_SHUFFLE)?-2:-3);
	  break;
	case ']':
	  play_song((playmode==_PLAYMODE_SHUFFLE)?-2:-4);
	  break;
        case '}':
          play_song((playmode==_PLAYMODE_SHUFFLE)?-2:-6);
          break;
        case '{':
          play_song((playmode==_PLAYMODE_SHUFFLE)?-2:-5);
          break;
    	case '>':
          hopseek(15);
    	  break;
    	case '<':
          hopseek(-15);
    	  break;
    	case '1':  /* #'s 1-9 */
    	case '2':
    	case '3':
    	case '4':
    	case '5':
    	case '6':
    	case '7':
    	case '8':
    	case '9':
    	  play_song(ch-'0');
    	  break;
    	case 'a': play_song(0xa); break;
    	case 'b': play_song(0xb); break;
    	case 'c': play_song(0xc); break;
    	case 'd': play_song(0xd); break;
    	case 'e': play_song(0xe); break;
    	case 'f': play_song(0xf); break;
#ifdef OLD_NCURSES
        case 7: /* 16-36 (hexa-tri) */
        case 8:
        case 9:
        case 10:
        case 11:
        case 12:
        case 13:
        case 14:
        case 15:
        case 16:
        case 17:
        case 18:
        case 19:
        case 20:
        case 21:
        case 22:
        case 23:
        case 24:
        case 25:
        case 26:
#else
#endif        
          play_song(ch+9);
          break;
    	case 'm':
    	  playmode++;
    	  if (playmode>_PLAYMODE_DISCLOOP)
    	    playmode=_PLAYMODE_NORM;
    	  if (playmode==_PLAYMODE_TRACKLOOP)
    	    looped_track=cur_track;
    	  else if (playmode==_PLAYMODE_SHUFFLE) {
    	    looped_track=-1;
    	    lasttrack=cur_track;
    	  }
    	  show_mode();
   		  break;
	case 'r':
/* #ifdefs put inside the case statements because I so often
 * forget which ones I've used already
 */
#ifdef SQL_DBASE
/*          report_status("DB REFRESH");*/
          sql_lookup_cd(thiscd.nid,&thiscd);
/*          draw_tracklist();*/
#endif	  
          break;
        case 't':
          read_local_dbase(thiscd.nid,&thiscd);
          update_topline();
          break;

	case 'u':
#ifdef SQL_DBASE_SUBMIT
#if DEBUGLEVEL>1
          f_write(1,LINES-1,10,"presubmit");
#endif
	  sql_submit_cd(thiscd.nid,&thiscd);
#endif
          break;
        case 'w':
          write_dbase();
          break;
        case 'n':
          if (listmode==_LISTMODE_EXT)
            break;
          ccc=~ccc;
          if (ccc)
            ccc_draw(-1);
          else
            draw_menu(menunum);
          break;
    	case 'j':
    	  if (playing)
    	    break;
    	  cd_lock(UNLOCK);
    	  f_write(19,16,5," ");
    	  cd_cmd(EJECT_TRAY);
#ifdef XTERM_TLINE
          if (!strncmp((char *)getenv("TERM"),"xterm",5))
            printf("\033]0;lyn - no disc\007");
#endif
/*
    	  report_status("NO DISC");
          blank_tracklist();
          diskin=0;*/
          break;
/*
        case 'k':
    	  if (playing)
            break;
          dwrite(11,5,15,"CLOSING");
		  cd_cmd(CLOSE_TRAY);
		  disklen=cdrom_data.endofdisk;
		  error_ctr=0;
          dwrite(11,5,15,"LOGGING");
		  do {
		    cd_get_audio_info();
		    error_ctr++;
            if (kbhit()) {
              ch=toupper(getch());
              if (ch=='Q' || ch=='X' || ch==27)
                done=1;
	        }
            if (cdrom_data.error==0x8102 && !done) {
	          usleep(SLEEP_USECS);
#if DEBUGLEVEL>2
              dwrite(21,2,15,"(Sub-K: %d)",error_ctr);
#endif
            }
		  } while (cdrom_data.error==0x8102 && error_ctr<135 && !done);
		  if (error_ctr==150)
		    nodisk();
		  if (cdrom_data.endofdisk!=disklen)
		    newdisk();
		  break;*/
/*        case 'L':
          cd_status();
          oldstatus=cdrom_data.status;
          if (cdrom_data.status & 2) {
            cd_lock(LOCK);
            dwrite(19,16,5,"");
          } else {
	        cd_lock(UNLOCK);
            dwrite(19,16,5," ");
	      }
	      break;*/
        case 'v':
          vol=100;
          /*cd_volume(255,bal,255);*/
          SETVOL(vol,vol);
          DRAWVOL();
	  break;
        case '0':
          vol=0;
          /*cd_volume(vol,bal,255);*/
          SETVOL(vol,vol);
          DRAWVOL();
	  break;
        case '+':
          if (vol<100)
            vol++;
          /*cd_volume(vol,bal,255);*/
          SETVOL(vol,vol);
          DRAWVOL();
          break;
        case '!': vol=10; SETVOL(vol,vol); DRAWVOL(); break;
        case '@': vol=20; SETVOL(vol,vol); DRAWVOL(); break;        
        case '#': vol=30; SETVOL(vol,vol); DRAWVOL(); break;
        case '$': vol=40; SETVOL(vol,vol); DRAWVOL(); break;
        case '%': vol=50; SETVOL(vol,vol); DRAWVOL(); break;
        case '^': vol=60; SETVOL(vol,vol); DRAWVOL(); break;
        case '&': vol=70; SETVOL(vol,vol); DRAWVOL(); break;
        case '*': vol=80; SETVOL(vol,vol); DRAWVOL(); break;
        case '(': vol=90; SETVOL(vol,vol); DRAWVOL(); break;
        case ')': vol=100; SETVOL(vol,vol); DRAWVOL(); break;
        case '-': 
          if (vol)
            vol--;
          SETVOL(vol,vol);
          DRAWVOL();
          break;
        case 'x':
          stop_cd();
          dwrite((listmode==_LISTMODE_EXT)?30:64,6+cur_track,7,"%2d:",cur_track);
          f_write(19,16,5," ");
          playing=0;
          f_write(19,7,13,"   ");
          f_write(19,8,13,"      ");
          f_write(19,9,13,"      ");
          report_status("STOPPED");
        case 'q':
        case 27:
          done=1;
          break;
      }
    } else
      usleep(CYCLE_SLEEP_USECS);
    if (playing)
      if (!paused && cd_done_play() && playmode==_PLAYMODE_DISCLOOP)
	play_song(cdrom_data.low_audio);
      else if (cur_cdmode==4) {
	playing=0;
	f_write(19,16,5," ");
	f_write(19,7,13,"   ");
	f_write(19,8,13,"      ");
	f_write(19,9,13,"      ");
	if (playmode==_PLAYMODE_SHUFFLE) {
	  f_write(11,5,15,"SHUFFLE");
	  play_song(-2);
	} else
#if DEBUGLEVEL>2
	  if (no_disc)
	    report_status("NO DISC");
	  else
#endif
            report_status("STOPPED");
      } else {
        where_the_fuck_are_we();
      }
  }
  attrset(COLOR_PAIR(15));
  for (i=0; i<=LINES; i++)
    f_write(1,i,15,charstr(COLS,' '));
  mix_close();
  endwin();
  return 0;
#pragma warn -aus
}
#pragma warn .aus

void where_the_fuck_are_we(void)
{
  static int min,sec,amin,asec,tsec,csec;
#ifdef XTERM_TLINE
  static time_t last_xtline=0;
#endif

  tsec=cur_frame/75;
  csec=cur_pos_rel;
  sec=csec % 60;
  min=(csec-sec)/60;
  asec=tsec % 60;
  amin=(tsec-asec)/60;
  cd_status();
  if ((cur_track!=lasttrack) &&
      (cur_track<=thiscd.ntracks) &&
      (cur_track>=1)) {
/*    highlight_track(0,cur_track);
    cur_track=cur_track;*/
    highlight_track(1,cur_track);
  }
  dwrite(19,7,13,"%d   ",cur_track);
  dwrite(19,8,13,"%02d:%02d",min,sec);
  dwrite(19,9,13,"%02d:%02d",amin,asec);
#ifdef XTERM_TLINE
  if (!strncmp((char *)getenv("TERM"),"xterm",5)) {
    printf("\033]0;lyn - %s, %s",thiscd.title,thiscd.songtitle[cur_track-1]);
    if (playing)
      printf(" (%02d:%02d)",min,sec);
    printf("\007");
  }
#endif
/*  update_xtermtitle(playing,min,sec); */
}


void ccc_draw(int cur)
{
  int i;

  if (cur==-1) {
    for (i=0; i<LINES-6; i++)
      f_write(27,5+i,8,charstr(24,32));
    f_write(27,5,15,"DEBUG DATA");
    f_write(31,6,15,  "C#  Val  %%");
  }
  cwrite(31,7,7, "LV:"); /*|10 %-3.3d|2  %.1f%%  ",(float)((20-bal)/20.)*vol,(float)(((20-bal)/20.)*vol)/2.55);*/
  cwrite(31,8,7, "RV:"); /*|10 %-3.3d|2  %.1f%%  ",(float)((bal)/20.)*vol,(float)(((bal)/20.)*vol)/2.55);


  cwrite(31,10,7,"Bal:|10 %d/%d",20-bal,bal);
  cwrite(31,12,7,"Read:|10 ioctl");
  cwrite(31,13,7,"ID  :|10 %s",thiscd.cid);
/*  cwrite(31,14,7,"PID :|10 _%d",getpid());*/
  cwrite(31,15,7,"Index :|10 0x%x",cur_index);
  cwrite(31,16,7,"Status:|10 0x%x",cur_cdmode);
  cwrite(31,17,7,"Frames:|10 0x%x",thiscd.length*75);
  cwrite(31,18,7,"CurLoc:|10 0x%x",cur_frame);
  cwrite(31,19,7,"CurPos:|10 0x%x",cur_pos_abs);
  cwrite(31,20,7,"Length:|10 0x%x",thiscd.length);
  cwrite(31,22,7,"HasColor:|10 %d",has_colors());
  if (has_colors() && LINES>26) {
    cwrite(31,23,7,"ChgColor:|10 %d",can_change_color());
    cwrite(31,24,7,"ClrPairs:|10 %d",COLOR_PAIRS);
    cwrite(31,25,7,"TermUsed:|10 %s",getenv("TERM"));
  } else {
    cwrite(31,23,7,"TermUsed:|10 %s",getenv("TERM"));
  }
  cwrite(31,24,7,"CDDB ID:|10 %s",thiscd.cddb_id);
/*  cwrite(31,25,7,"ArrChars:|10 %d %d %d %d",ACS_LARROW,ACS_RARROW,ACS_UARROW,ACS_DARROW);*/
/*  for (i=0; i<8; i++) {
    dwrite(31,25+i,7,"Color %-2d:",i);
    cwrite(41,25+i,i,"%d|%d %d",i,i+8,i+8);
  }*/
}


void nodisk(void)
{
#if DEBUGLEVEL
  no_disc=1;
#endif
  playing=diskin=0;
  f_write(19,14,13,"none ");
  blank_tracklist();
  blank_trackdata();
  report_status("NO DISC");
  report_topline(8,copyright_string);
}

void highlight_track(int mode, unsigned int track)
{
  char s[81];
  int ctr;

  if (!track || track>cur_track || _fast)
    return;
  for (ctr=tracklist_top; ctr<=tracklist_bottom; ctr++) {
    dwrite(LISTMODE_LEFTX+2,7+ctr-tracklist_top,(ctr==track && mode==1)?31:7,"%c%2d%c",pl.used?'+':32,ctr,pl.used?32:':');
    if (listmode==_LISTMODE_EXT)
      f_write(41,7+ctr-tracklist_top,(mode==1 && ctr==track)?15:7,thiscd.songtitle[ctr-1]);
  }
  if (thiscd.title[0]!='<' && thiscd.title[0] && listmode==_LISTMODE_NORM) {
    if (playing)
      sprintf(s,"%s, %s",thiscd.title,thiscd.songtitle[cur_track-1]);
    else
      strcpy(s,thiscd.title);
    report_topline(15,s);
  }
}

void draw_tracklist(void)
{
  int i,min,sec;
  char *s;
  

  if (listmode==_LISTMODE_EXT)
    for (i=0; i<LINES-3; i++)
      f_write(27,4+i,8,charstr(COLS-26,' '));
  else
    for (i=0; i<LINES-7; i++)
      f_write(61,7+i,8,charstr(COLS-60,' '));
  if (listmode==_LISTMODE_EXT) {
    cstr(27,4,8,COLS-26);
    s=(char *)malloc(strlen(thiscd.title));
    strcpy(s,thiscd.title);
    if (strlen(s)>COLS-40) {
      s[COLS-43]=0;
      strcat(s,"...");
    }
    dwrite(27,5,15,"SONGLIST    %s",s); 
    f_write(36,5,7,"::");
    f_write(30,6,15,"##  Len    Song");
    free(s);
  }
  for (i=tracklist_top; i<=tracklist_bottom; i++) {
    sec=thiscd.trk[i-1].length % 60;
    min=(thiscd.trk[i-1].length-sec)/60;
    if (listmode==_LISTMODE_EXT)
      cwrite(29,7+i-tracklist_top,7,"%c%2d%c|11 %02.02i:%02i|%d  %-42.42s",pl.used?'+':32,i,pl.used?32:':',
        min,sec,7,thiscd.songtitle[i-1]);
    else
      cwrite(64,7+i-tracklist_top,7,"%2d:|11 %02.02i:%02.2i",i,min,sec);
/*    if (pl.used)
      cwrite((listmode==_LISTMODE_EXT)?29:63,7+i-tracklist_top,7,"+%-3d",pl.entry[i]);*/
    if (listmode==_LISTMODE_NORM)
      if (thiscd.trk[i-1].data) {
        f_write(75,7+i-tracklist_top,3,"Data  ");
        f_write(68,7+i-tracklist_top,8,"       ");
      } else
        f_write(75,7+i-tracklist_top,3,"Audio ");     
    else
      if (thiscd.trk[i-1].data)
        f_write(34,7+i-tracklist_top,8,"       ");
  }
  if (i==tracklist_top) {
/*    dwrite(61,7,8,charstr((listmode==_LISTMODE_EXT)?COLS-27:COLS-60,LCHAR)); */
    cstr((listmode==_LISTMODE_EXT)?27:61,7,8,(listmode==_LISTMODE_EXT)?COLS-27:COLS-60);
    return;
  }
  sec=thiscd.length % 60;
  min=(thiscd.length-sec)/60;
  cwrite((listmode==_LISTMODE_EXT)?27:61,i-tracklist_top+7,7,"Total:|11 %02d:%02d",
    min,sec);
  cstr((listmode==_LISTMODE_EXT)?27:61,i-tracklist_top+8,8,(listmode==_LISTMODE_EXT)?COLS-27:COLS-60);
  
/*  if (playing)*/
  highlight_track(1,cur_track);
  refresh();
}

void newdisk(void)
{
  char s[81],s1[81];
  short error_ctr,i;

#ifdef XTERM_TLINE
  if (!strncmp((char *)getenv("TERM"),"xterm",5))
    printf("\033]0;lyn - logging\007");
#endif
  cd_status();
  playmode=_PLAYMODE_NORM;
  error_ctr=0;
  diskin=1;
  report_status("LOGGING");
  num_tracks=thiscd.ntracks; 
/*  memset(&s1,0,81); */
  sprintf(s1,"%ld",(long)(thiscd.trk[0].length*thiscd.ntracks*thiscd.length));
  thiscd.nid=thiscd.trk[0].length*thiscd.ntracks*thiscd.length;
  if (thiscd.nid<0)
    thiscd.nid*=-1; 
/*  if (thiscd.nid!=lastid) { */
    strcpy(thiscd.cid,s1);
    memset(&thiscd.cddb_id,0,sizeof(thiscd.cddb_id));
    gen_cddb_id();
    if (read_dbase()==-1)
      thiscd.nid=0;
/*  }*/
  lastid=thiscd.nid;
 if (!thiscd.nid)
/*    blank_trackdata()*/;
  else if (listmode==_LISTMODE_NORM)
    report_topline(15,thiscd.title);
  for (i=0; i<thiscd.ntracks; i++)
    pl.entry[i]=i+1;
  pl.used=0;
  pl.length=thiscd.length;
  pl.tsec=pl.length % 60;
  pl.tmin=(pl.length-pl.tsec)/60;
  if (playing) {
    if (thiscd.ntracks>LINES-8) {
      if (cur_track>16) {
        tracklist_top=cur_track-(LINES-8);
        tracklist_bottom=cur_track;
#if DEBUGLEVEL>4
        dwrite(1,LINES-1,15,"tt=%d tb=%d cb=%d",
          tracklist_top,tracklist_bottom,cur_track);
#endif
      } else {
        tracklist_top=1;
        tracklist_bottom=LINES-8;
      }
      pl.cpt=cur_track;
    } else {
      tracklist_top=1;
      tracklist_bottom=thiscd.ntracks;
    }
  } else {
    tracklist_top=1;
    tracklist_bottom=thiscd.ntracks>=(LINES-8)?(LINES-8):thiscd.ntracks;
  }
  f_write(61,8,7,charstr(19,32));
  f_write(61,7,7,charstr(19,32));
  f_write(19,12,13,last_token('/',config.active_cdromdevice));
  draw_tracklist();
#ifdef XTERM_TLINE
  if (!strncmp((char *)getenv("TERM"),"xterm",5)) {
    printf("\033]0;lyn - %s, %s",thiscd.title,thiscd.songtitle[cur_track-1]);
    if (playing)
      printf(" (%02d:%02d)",(cur_pos_rel-(cur_pos_rel % 60))/60,
                            cur_pos_rel % 60);
    printf("\007");
  }
#endif
  DRAWVOL();
  if (!playing) {
    cd_set_track(1);
    report_status("STOPPED");
  } else
    report_status("PLAYING");
  show_mode();
  show_status();
  disklen=thiscd.length;
}


short play_song(int num)
{
  short temp,newt,tt;
  unsigned long startpos, endpos;

  curcheck=1;
  if (num==-1) /* first track */
    cd_set_track(1);
  else if (num==-2) { /* random track */
    cd_stop_audio();
    srand(time(NULL)+loopctr);
    newt=tt=(rand() % thiscd.ntracks)+1;
    cd_set_track(newt);
/*
    lastrand=loopctr;
    startpos=thiscd.trk[newt-1].start;
    endpos=thiscd.trk[newt].start+(thiscd.trk[newt].length*75);
    play_chunk(startpos,endpos);
    cd_seek(cdrom_data.track_position);
    playing=1;
    if (!_fast) {
      highlight_track(1,cur_track);
      report_status("PLAYING");
    }
    lasttrack=cur_track;
    usleep(400000);
    return 1;*/
  } else if (num==-3) { /* previous track */
    temp=tt=cur_track;
    if (temp>1) {
      tt--;
      unhighlight_tracknums();
/*      cd_stop_audio();
      cd_set_track(temp-1);*/
      if (temp-1<tracklist_top) {
	tracklist_top--;
	tracklist_bottom--;
	draw_tracklist();
	highlight_track(1,temp-1);
      } else
	highlight_track(1,temp-1);
    }
  } else if (num==-4) { /* next track */
    temp=tt=cur_track;
    if (temp<thiscd.ntracks) {
      tt++;
/*      unhighlight_tracknums();
      cd_stop_audio();
      cd_set_track(temp+1);*/
      if (temp+1>tracklist_bottom) {
	tracklist_top++;
	tracklist_bottom++;
	draw_tracklist();
	highlight_track(1,temp+1);
      } else
	highlight_track(1,temp+1);
    }
  } else if (num==-5) { /* back 5 tracks */
    temp=tt=cur_track;
    if (temp-4>1) {
      tt-=4;
      unhighlight_tracknums();
      cd_stop_audio();
      cd_set_track(temp-5);
      if (temp-5<tracklist_top) {
	tracklist_top-=5;
	tracklist_bottom-=5;
	draw_tracklist();
	highlight_track(1,temp-5);
      } else
	highlight_track(1,temp-5);
    }
  } else if (num==-6) { /* up five tracks */
    temp=tt=cur_track;
    if (temp+4<thiscd.ntracks) {
      tt+=5;
      if (!_fast)
	unhighlight_tracknums();
      cd_stop_audio();
      cd_set_track(temp+5);
      if (temp+5>tracklist_bottom) {
	tracklist_top+=5;
	tracklist_bottom+=5;
	draw_tracklist();
	highlight_track(1,temp+5);
      } else
	highlight_track(1,temp+5);
    }
  } else if ((num<1) || (num>thiscd.ntracks))
    return 0;
  if (playing && !_fast)
    unhighlight_tracknums();
  if (num>0) {
    tt=num;
    cd_stop_audio();
    if (num<tracklist_top) {
      tracklist_top=num;
      tracklist_bottom=thiscd.ntracks>LINES-8?tracklist_top+(LINES-9):thiscd.ntracks;
      draw_tracklist();
    } else {
      unhighlight_tracknums();
      highlight_track(1,num);
    }
    cd_set_track(num);
  }
  startpos=thiscd.trk[tt-1].start;
  if (playmode==_PLAYMODE_NORM)
    endpos=thiscd.length*75;
  else {
    temp=cur_track;
    if ((temp+1)>thiscd.ntracks)
      endpos=thiscd.trk[temp].start+(thiscd.trk[temp].length*75);
    else
      endpos=thiscd.trk[temp+1].start+(thiscd.trk[temp].length*75);
  }
  cd_seek(cdrom_data.track_position);
  play_chunk(startpos, endpos);
  playing=1;
  if (!_fast) {
    highlight_track(1,cur_track);
    report_status("PLAYING");
/*     dwrite(11,5,15,"PLAYING");*/
  }
  lasttrack=cur_track;
  usleep(400000);
  return 1;
}

void hopseek(signed int seconds)
{
  if (!seconds)
    return;
  cd_play_audio(cur_frame+(seconds*75),thiscd.length*75);
  where_the_fuck_are_we();
}

void pause_song(void)
{
  pause_cd();
  if (!paused) {
    paused=1;
    if (!_fast)
      report_status("PAUSED");
  } else {
    paused=0;
    if (!_fast)
      report_status("PLAYING");
  }
}








void cd_lock(int d)
{
}

void cd_set_track(short tracknum)
{
}

void cd_seek(unsigned long location)
{
}

short cd_done_play(void)
{
  int i=cd_status();
  
  if (!i || i==4)
    return 1;
  return 0;
}



